"""
Create simple csv files to test joining tables.
"""
from astropy.table import Table
import numpy as np

# Table with single parameters and three references to test
names = ["pl_name", "mypar", "myparerr1", "myparerr2", "mypar_reflink", "default_flag"]
data1 = [["Cool-planet b", 0.5, 0.01, 0.01, "myref1", 1], ["Other-planet c", np.nan, np.nan, np.nan, "", 1]]
data2 = [["Cool-planet b", 0.8, 0.1, 0.1, "myref2", 0], ["Other-planet c", 0.2, 0.02, 0.02, "myref2", 0]]
data3 = [["Cool-planet b", 0.9, 0.2, 0.2, "myref3", 0], ["Other-planet c", 0.5, 0.2, 0.2, "myref3", 0]]

data_rows = [*data1, *data2, *data3]

tbl = Table(rows=data_rows, names=names)
tbl.write("tests/data/single_param.csv", overwrite=True)

# Add second parameter
tbl["otherpar"] = tbl["mypar"] * 4
tbl["otherparerr1"] = tbl["myparerr1"] * 4
tbl["otherparerr2"] = tbl["myparerr2"] * 4
tbl["otherpar_reflink"] = tbl["mypar_reflink"]

# Swap so that second parameters has NaN in different (non-default) refs
# Seems like need to look otherwise astropy gives copies of row sections
other_cols = ["otherpar", "otherparerr1", "otherparerr2", "otherpar_reflink"]
for oc in other_cols:
    r1_old = tbl[oc].copy()[1]
    r2_old = tbl[oc].copy()[3]
    tbl[oc][1] = r2_old
    tbl[oc][3] = r1_old
tbl["otherpar_reflink"][1] = "myref1"

tbl.write("tests/data/two_params.csv", overwrite=True)



from astropy.units import UnitsWarning


class ColUnitsWarning(UnitsWarning):

    def __init__(self, col, units):

        message = "Units conflict for '{}' column."  \
                + " Converting '{}' to '{}'."  \
                + " Make sure units were properly converted."
        message = message.format(col, *units)

        super().__init__(message)


class NoUnitsWarning(UnitsWarning):

    def __init__(self, col, units):

        message = "Units conflict for '{}' column."  \
                + " No units were specified."  \
                + " Assuming '{}'."
        message = message.format(col, units)

        super().__init__(message)


class MultipleResultsWarning(UserWarning):

    def __init__(self, input_value, results):

        message = f"Multiple values where found for {input_value}: "
        message += ', '.join([f'{name}' for name in results])

        super().__init__(message)


class NotFoundWarning(UserWarning):

    def __init__(self, *input_values):

        message = f"No value was found for "
        message += ', '.join([f'{name}' for name in input_values])

        super().__init__(message)


class NotFoundError(ValueError):

    def __init__(self, *input_values):

        message = f"No value or multiple values were found for: "
        message += ', '.join([f'{name}' for name in input_values])

        super().__init__(message)


class BaseFileWarning(UserWarning):

    def __init__(self, message, file=None, err=None):

        if message is None:
             message = ""

        if err is None:
            err = 'UnspecifiedException'
        else:
            err = err.__class__.__name__

        if file is None:
            file = 'unspecified file'

        # Print error name and file in message
        message = message.format(file.upper(), err, file)

        super().__init__(message)


class GetLocalFileWarning(BaseFileWarning):

    def __init__(self, message=None, file=None, err=None):

        if message is None:
             message = "DID NOT READ {}. {} has occur "  \
               + "when trying to query/read {}."

        super().__init__(message, file=file, err=err)


class QueryFileWarning(BaseFileWarning):

    def __init__(self, message=None, file=None, err=None):

        if message is None:
            message = "QUERY {} FAILED. {} has occur "  \
                    + "when trying to query {}."

        super().__init__(message, file=file, err=err)


from pathlib import Path
from exofile.archive import ExoFile

# PATH_TO_EXODIR = Path("/home/adb/www/")
PATH_TO_EXODIR = Path(".")
PATH_TO_FILE = PATH_TO_EXODIR / "exofile.ecsv"
PATH_TO_FILE_ALT = PATH_TO_EXODIR / "exofile_alt.ecsv"

new = ExoFile.update()
new.write(PATH_TO_FILE, overwrite=True)
new_alt = ExoFile.update(use_composite_archive=False)
new_alt.write(PATH_TO_FILE_ALT, overwrite=True)



from warnings import warn

import numpy as np
from astropy.uncertainty.core import NdarrayDistribution

from exofile.archive import load_exoplanet_archive_mappings, ExoArchive, difference
from exofile.table_custom import MaskedColumn


def migrate_table(old_tbl, old_ref):

    label_df = load_exoplanet_archive_mappings()
    label_df_all = label_df.copy()
    NEW_NAME = "Planetary Systems Composite Parameters (PSCP)"
    OLD_NAME = "Confirmed Planets (retiring)"
    label_df = label_df.dropna(subset=[NEW_NAME, OLD_NAME])
    label_df = label_df[~label_df[NEW_NAME].str.contains("systemref")]
    label_new = label_df[NEW_NAME].str.strip()
    label_old = label_df[OLD_NAME].str.strip()

    for lold, lnew in zip(label_old, label_new):
        # Some columns are missing even if in CSV, so will have a few warnings
        try:
            old_tbl.rename_column(lold, lnew)
            old_ref.rename_column(lold, lnew)
        except KeyError:
            warn(f"Column {lold} was not found in old table", RuntimeWarning)

    pc_struct = ExoArchive.query(select="top 1 *")
    pc_cols = pc_struct.colnames

    # Now remove extra columns
    dlist = difference(old_tbl.colnames, pc_cols)
    del old_tbl[dlist]
    del old_ref[dlist]

    missing_cols = difference(pc_cols, old_tbl.colnames)

    # Get ref an non-ref missing columns
    reflink_labels = label_df_all.dropna(subset=NEW_NAME)[NEW_NAME]
    reflink_labels = reflink_labels.str.strip()
    reflink_labels = reflink_labels[reflink_labels.str.endswith("reflink")]
    reflink_labels = reflink_labels[reflink_labels.isin(missing_cols)]
    non_ref_cols = difference(missing_cols, reflink_labels.to_list())

    # Add missing columns
    for col in non_ref_cols:
        if pc_struct[col].dtype.kind in ["U", "S"]:
            old_tbl[col] = ""
        else:
            old_tbl[col] = MaskedColumn(
                np.full(len(old_tbl), np.nan), unit=pc_struct[col].unit
            )

    for rlab in reflink_labels:

        plab = rlab.replace("_reflink", "")

        # If column was not in old file, it has no refs
        if plab in non_ref_cols:
            new_refs = ""
        else:
            new_refs = old_ref[plab]

        old_tbl[rlab] = MaskedColumn(new_refs)

    new_tbl = old_tbl[pc_cols]

    return new_tbl


def get_tp_from_tau(
        tau, per, tref, return_dist=False
    ):
    """
    Get time of periastron from period, "tau" parameter and a reference time.

    The tau parameter, sometimes labeled "chi", is tau = (tref - tp) / per
    where tref is a reference time (usually the first obseravtion time).

    This function is provided as a utility because many references provide only tau
    and exofile does not support this parameter (nor does the NASA exoplanet archive).

    :param tau: Periastron reference parameter
    :type tau: Union[NdarrayDistribution, float]
    :param per: Orbit period
    :type per: Union[NdarrayDistribution, float]
    :param tref: Reference time to calculate tp
    :type tref: float
    :param return_dist: Whether an astropy distribution should be returned, defaults to False
    :type return_dist: bool, optional
    """

    has_err = isinstance(tau, NdarrayDistribution) or isinstance(per, NdarrayDistribution)

    tp = tref - tau * per

    if return_dist and has_err:
        return tp
    elif return_dist:
        warn(
            "Cannot return a distribution because inputs are not astropy distributions."
            " Returning a scalar."
        )
        return tp
    elif has_err:
        return (tp.pdf_mean(), tp.pdf_std())
    else:
        return tp



"""
Helper to migrate the old exofile format to the new one.
Most of this is taken from the ExoFile.update() and adapted to the old->new conversion
instead of PS -> PSCP.
"""

from exofile.archive import ExoFile
from exofile.utils import migrate_table

NEW_PATH = "exofile.ecsv"

old_tbl = ExoFile.read("old_exofile.ecsv")
old_ref = ExoFile.read("reference_file.ecsv")

new_tbl = migrate_table(old_tbl, old_ref)

new_tbl.write(NEW_PATH)



import re
from warnings import warn
from pathlib import Path
from typing import Any, List, Optional, Tuple, Union

import numpy as np
import requests
from astropy.time import Time
from astropy.units import Unit
from astroquery.ipac.nexsci.nasa_exoplanet_archive import NasaExoplanetArchive
from pandas import read_csv, DataFrame

from .config import Param
from .exceptions import (ColUnitsWarning, GetLocalFileWarning, NoUnitsWarning,
                         QueryFileWarning)
from .table_custom import MaskedColumn, Table, difference


# Columns that are missing in archive even though listed in master CSV mappign
MISSING_COLS = [
    "raerr1",
    "raerr2",
    "decerr1",
    "decerr2",
    "glaterr1",
    "glaterr2",
    "glonerr1",
    "glonerr2",
    "elaterr1",
    "elaterr2",
    "elonerr1",
    "elonerr2",
]
PS_NAME = "Planetary Systems (PS)"
PC_NAME = "Planetary Systems Composite Parameters (PSCP)"
ARCHIVE_CSV_PATH = "https://exoplanetarchive.ipac.caltech.edu/docs/Exoplanet_Archive_Column_Mapping_CSV.csv"

def get_refname_from_link(link: str) -> str:
    """
    Get the text from a reflink
    """
    if link != "":
        res = re.search(">(.*)</a>", link)
        if res is not None:
            out = res.group(1)
        else:
            warn(f"Regex result for link {link} is None, returning input link")
            out = link
    else:
        out = link

    return out


def get_refname_from_links(links: List[str]) -> List[str]:
    """
    Get the text from multiple reflinks
    """
    return [get_refname_from_link(link) for link in links]


class ExoFile(Table):

    main_col = "pl_name"

    def get_colnames_with_error(
            self, err_ext: List[str] = ["err1", "err2"]
        ) -> Tuple[str]:
        """
        Find all columns with an error column associate`.
        `err_ext` is the list of extensions for the name of the error columns.
        Returns a tuple of lists, each lists contain the name of a column
        and the names of its related errors.

        Example:
        colname = 'pl_tperi'
        errcols: 'pl_tperierr1' and 'pl_tperierr2'
        The output would be: (['pl_tperi', 'pl_tperierr1', 'pl_tperierr2'], )
        """

        colnames = ()
        for key in self.colnames:

            try:
                for err in err_ext:
                    self[key + err]
                colnames += ([key + err for err in ["", *err_ext]],)

            except KeyError:
                pass

        return colnames

    def mask_no_errors(self, **kwargs):
        """
        Mask columns where errors are not available
        """
        for cols in self.get_colnames_with_error(**kwargs):

            mask = self[cols].mask
            # Convert to regular array
            mask = np.array(mask.as_array().tolist())

            # Find where values, err1 or err2 are masked
            imask = np.where(mask.any(axis=-1))

            # Mask values, err1 and err2 at these rows
            for col in cols:
                self[col].mask[imask] = True

    def mask_zero_errors(self, **kwargs):
        """
        Mask columns where associated errors are equal to zero
        """
        for cols in self.get_colnames_with_error(**kwargs):

            # Put all errors in a 2d array (ex: err1, err2)
            errors = [self[col].data for col in cols[1:]]
            errors = np.ma.array(errors).T

            # Find where err1 or err2 are zero
            cond = errors == 0.0
            imask = np.where(cond.any(axis=-1))[0]

            # Mask value, err1 and err2 at these rows
            for col in cols:
                if imask.any():
                    self[col].mask[imask] = True

    def replace_with(self, other, main_col=None, warn_units=True):
        """
        Use all non-masked values of `other` to replace values in `self`.
        `key` is the column use to identify corresponding rows.
        Default is other.main_col
        """

        # Take main_col if key is not given
        main_col = main_col or other.main_col

        # Make sure `other` is masked
        other = ExoFile(other, copy=True, masked=True)

        # Save units for conversion
        units = {}
        for key in other.keys():
            units[key] = self[key].unit
            # Check if they are the same
            if other[key].unit != units[key] and warn_units:
                if other[key].unit is None:
                    other[key].unit = units[key]
                    warn(NoUnitsWarning(key, units[key]))
                else:
                    warn(ColUnitsWarning(key, [other[key].unit, units[key]]))

        # Get position in main table
        index = self.get_index(*other[main_col], name_key=main_col)
        index = np.array(index)

        # Add empty rows for objects not in `self`
        for name in other[index == -1]["pl_name"]:
            self.add_row({"pl_name": name})

        # Get position in main table again
        index = self.get_index(*other[main_col], name_key=main_col)
        index = np.array(index)

        # Replace values of objects in `self`
        for i_other, i_self in enumerate(index):
            # Position of keys that are not masked
            (i_keys,) = np.where(~np.array(list(other.mask[i_other])))

            # Keys not masked
            keys = [other.keys()[i_key] for i_key in i_keys]

            # Don't change values for main column
            keys.remove(main_col)

            # Assign new values
            for key in keys:
                # Don't convert units if unit is None. Else do so.
                if units[key] is None:
                    self[i_self][key] = other[i_other][key]
                else:
                    conversion = other[key].unit.to(units[key])
                    self[i_self][key] = other[i_other][key] * conversion

    @classmethod
    def query(
            cls,
            url_key="url",
            debug=False,
            exofile_kwargs=None,
            use_alt_file=False,
            warn_units=True,
            warn_local_file=True,
            **kwargs
        ):
        """
        Query the exofile and try to complement it with
        the custom table (google sheet).
        Returns the complemented exofile.
        Parameters:
        - url: string
            adress of the exofile
        - url_key: string
            key to get the url from param file. Default is 'url'.
        - sheet_key: string
            Identification key of the google sheet
        - use_alt_file: bool
            Use alternative exofile with the row-by-row composite table
        - warn_units: bool
            Warn about unit conversions when replace values from custom file
        - warn_local_file: bool
            Warn about missing local file (set to false if not local file is expected)
        """
        # Get links to query the different database
        param = Param.load().value
        # Use input arguments if given
        param = {**param, **kwargs.pop("param", {})}

        alt_str = "_alt" if use_alt_file else ""

        try:
            # Get combined table
            master = requests.get(param[url_key])
            master.raise_for_status()
            # Convert to table
            master = cls.read(master.text, format="ascii")
        except (requests.exceptions.SSLError, requests.exceptions.HTTPError):
            # If SSLError or HTTPError, maybe just the exofile website is problem,
            # still try local file and then google sheet for custom
            # Handling this here allows using the Google sheet even if online exofile
            # not available
            try:
                master = cls.read(param[f"exofile{alt_str}"], **exofile_kwargs)
            except Exception as e:
                if warn_local_file:
                    if debug == "raise":
                        raise e
                    warn(GetLocalFileWarning(file=f"exofile{alt_str}", err=e))

        # Try to complement with custom values
        try:
            # Get custom values to be added to the exofile
            custom = GoogleSheet.query(param["sheet_key"], **kwargs)
            # Replace values in the exofile
            master.replace_with(custom, warn_units=warn_units)

        except Exception as e:
            if debug:
                raise e
            warn(QueryFileWarning(file="google sheet", err=e))

        return master

    @classmethod
    def load(
        cls,
        query=True,
        param=None,
        debug=None,
        query_kwargs=None,
        exofile_kwargs=None,
        use_alt_file=False,
        warn_units=True,
        warn_local_file=True,
        **kwargs,
    ):
        """
        Returns the exofile complemented.
        Parameters
        - param: dict
            dictionnairy of the local files names. Default is param file.
            keys: 'exofile' and 'custom_file'
            `custom_file` will be use to replace values in the exofile.
            `exofile` is the local exofile.
        - query: bool
            query or not the exofile. If False, simply read `custom_file`
        - query_kwargs: None or dictionnary
            Passed to query() method
        - exofile_kwargs: None or dictionnary
            Passed to table.read() when reading the local exofile
            (see astropy.table.read)
        - use_alt_file: bool
            Use alternative exofile with the row-by-row composite table
        - warn_units: bool
            Warn about unit conversions when replace values from custom file
        - warn_local_file: bool
            Warn about missing local file (set to false if not local file is expected)
        - kwargs
            Passed to table.read() when reading the custom table
            (see astropy.table.read)
        """
        # Assign default dictionnary values
        #  Never put {} in the function definition
        if param is None:
            param = {}
        if query_kwargs is None:
            query_kwargs = {}
        if exofile_kwargs is None:
            exofile_kwargs = {}

        # Use module parameters and complete with input parameters
        param = {**Param.load().value, **param}

        alt_str = "_alt" if use_alt_file else ""

        ###########################
        # Complement the exofile
        ###########################
        # Query online if True
        if query:
            # Try to query the complemented exofile.
            # If impossible, set query to False
            try:
                master = cls.query(
                    **query_kwargs,
                    url_key=f"url{alt_str}",
                    exofile_kwargs=exofile_kwargs,
                    use_alt_file=use_alt_file,
                    warn_units=warn_units,
                    warn_local_file=warn_local_file,
                )
            except Exception as e:
                if debug == "raise":
                    raise e
                warn(QueryFileWarning(file=f"exofile{alt_str}", err=e))
                query = False

        # Read local exofile if query is False or query failed
        if not query:
            # Try to read locally
            try:
                master = cls.read(param[f"exofile{alt_str}"], **exofile_kwargs)
            except Exception as e:
                if warn_local_file:
                    if debug == "raise":
                        raise e
                    warn(GetLocalFileWarning(file=f"exofile{alt_str}", err=e))
                    # Return custom_file as last ressort
                    return cls.read(param["custom_file"], **kwargs)

        # Finally, complement with the local `custom_file`
        try:
            custom = cls.read(param["custom_file"], **kwargs)
            master.replace_with(custom, warn_units=warn_units)
        except Exception as e:
            if warn_local_file:
                if debug == "raise":
                    raise e
                warn(GetLocalFileWarning(file="custom file", err=e))

        return master

    def fill_custom_ref(self, default_ref="custom ref"):

        # Fill ref if not set by user
        for col in self.colnames:
            # Don't reflink the reflinks
            if "_reflink" in col:
                continue

            rname = f"{col}_reflink"
            if rname not in self.colnames and col != self.main_col:
                self[rname] = MaskedColumn(
                    data=[default_ref] * len(self), mask=self[col].mask
                )

    def write_to_custom(self, *args, default_ref="custom ref", **kwargs):

        self.fill_custom_ref(default_ref=default_ref)

        file = Param.load().value["custom_file"]
        self.write(file, *args, **kwargs)

    def estim_ephemeride_err(
        self, ephemeride="pl_tranmid", err_ext="err1", orbper_ext=None
    ):
        """
        Compute error on the ephemeride time as of today.
        Create and add the column 'today_{ephemeride}{err_ext}'.
        Default will be: 'today_pl_tranmiderr1'
        """
        # Take same err_ext for
        if orbper_ext is None:
            orbper_ext = err_ext

        # Get transit mid time and period
        tranmid = self[ephemeride].to_array(units="d")
        orbper = self["pl_orbper"].to_array(units="d")

        # Compute number of period since transit mid time
        n_period = (Time.now().jd - tranmid) / orbper

        # Get corresponding errors
        orbper_err = self["pl_orbper" + orbper_ext].to_array(units="d")
        tranmid_err = self[ephemeride + err_ext].to_array(units="d")

        # Compute error as of today
        error_today = tranmid_err + n_period * orbper_err

        # Save as a column
        col = MaskedColumn(error_today, unit="d", name=f"today_{ephemeride}{err_ext}")
        self.add_column(col)

    @staticmethod
    def update(
            sort_keys: Optional[Union[str, List[str]]] = None,
            verbose: bool = True,
            use_composite_archive: bool = True,
            local_table: Optional[Union[str, Path]] = None,
        ):
        """
        Returns an updated masterfile built with the NasaExoplanetArchive.

        Parameters
        ----------
        sort_keys : str or list of str
            The key(s) to order the table by (passed to `astropy.table.sort`).
            If None, use the column 'today_tranmid_err'.
        """
        # Default sort keys
        if sort_keys is None:
            sort_keys = ["today_pl_tranmiderr1", "today_pl_orbtpererr1"]

        # Read new database from exoplanet archive
        if use_composite_archive:
            tbl_id = "pscomppars"
            tbl_name = PC_NAME
        else:
            tbl_id = "ps"
            tbl_name = PS_NAME

        if verbose:
            print(f"Querying {tbl_name}...", end="")

        # TODO: Make sure OK with composite tbl kwd if keep this
        if local_table is not None:
            new = ExoArchive.read(local_table)
            new = ExoArchive.format_table(new)
        else:
            new = ExoArchive.query(table=tbl_id)

        if not use_composite_archive:
            new = format_ps_table(new, verbose=verbose)
            new = compose_from_ps(new, sort_keys, verbose=verbose)

            # Get actual pc columns to compare with what's left.
            # Using CSV from Archive deletes too many columns
            pc_cols = ExoArchive.query(select="top 1 *").colnames

            # Now remove extra columns
            dlist = difference(new.colnames, pc_cols)
            del new[dlist]

            # Index with same order as pc_cols
            new = new[pc_cols]

            # Safety check
            if len(difference(new.colnames, pc_cols)) > 0 or len(difference(pc_cols, new.colnames)) > 0:
                raise RuntimeError(
                    f"The formatted {tbl_id} table is incompatible with the masterfile/composite table format"
                )

        if verbose:
            print("Done")

        ref_cols = [cn for cn in new.colnames if cn.endswith("reflink")]
        for cname in ref_cols:
            new[cname] = get_refname_from_links(new[cname])

        return new


class GoogleSheet(ExoFile):

    url_root = "https://docs.google.com/spreadsheets/d/{}/gviz/tq?tqx=out:csv&sheet={}"

    @classmethod
    def query(cls, key, sheet_name=0, check_units="silent", keep_units=True):

        url = cls.url_root.format(key, sheet_name)

        data = requests.get(url)

        # Convert to astropy table
        table = cls.read(data.text, format="ascii")

        # Remove units from the column name
        # The structure is: "name [units]"
        for key in table.keys():
            # Split name and units
            name, unit = key.split(" [")
            unit = unit.split("]")[0]

            # Rename and assign units
            table.rename_column(key, name)
            if unit != "None" and keep_units:
                table[name].unit = Unit(unit, parse_strict=check_units)
            else:
                table[name].unit = None

        return table


class ExoArchive(ExoFile):

    @classmethod
    def query(
            cls,
            table: str = "pscomppars",
            **criteria,
        ):
        """
        Query the NASA exoplanet using astroquery and return in a MasterFile object.
        Returns the full table by default
        Accepts the same criteria asu
        `astroquery.ipac.nexsci.nasa_exoplanet_archive.NasaExoplanetArchive.query_criteria()`
        """
        # Query and get astropy table
        data = NasaExoplanetArchive.query_criteria(table=table, **criteria)

        # Covnert to custom table
        data = cls(data, masked=True, copy=False)

        # sky_coord is duplicate of RA and DEC and causes problems with units and masking
        if "sky_coord" in data.colnames:
            del data["sky_coord"]

        # Correct units
        # ???: Is this still required now that we use astroquery ?
        data.correct_units(verbose=False)

        # Mask where errors are not available
        # NOTE: Calling separately because if the default (err1,err2) gives a KeyError,
        # we can try with err separately
        data.mask_no_errors()  # err1 and err2 cols
        data.mask_no_errors(err_ext=["err"])  # err cols

        # Mask where errors are set to zero
        data.mask_zero_errors()  # err1 and err2 cols
        data.mask_zero_errors(err_ext=["err"])  # err cols

        return data

    @classmethod
    def format_table(cls, data):

        # Covnert to custom table
        data = cls(data, masked=True, copy=False)

        # sky_coord is duplicate of RA and DEC and causes problems with units and masking
        if "sky_coord" in data.colnames:
            del data["sky_coord"]

        # Correct units
        # ???: Is this still required now that we use astroquery ?
        data.correct_units(verbose=False)

        # Mask where errors are not available
        # NOTE: Calling separately because if the default (err1,err2) gives a KeyError,
        # we can try with err separately
        data.mask_no_errors()  # err1 and err2 cols
        data.mask_no_errors(err_ext=["err"])  # err cols

        # Mask where errors are set to zero
        data.mask_zero_errors()  # err1 and err2 cols
        data.mask_zero_errors(err_ext=["err"])  # err cols

        return data


def load_exoplanet_archive_mappings(
        csv_path: Optional[Union[str, Path]] = None
    ) -> DataFrame:

    csv_path = csv_path or ARCHIVE_CSV_PATH
    label_df = read_csv(
        csv_path,
        skiprows=[0, 2, 3],
        header=0,
        usecols=lambda x: not x.startswith("Unnamed: "),
    )

    return label_df


def format_ps_table(
        ps_tbl: Any,
        verbose: bool = False,
    ):
    # Map columns between PS and PS composite tables
    try:
        label_df = load_exoplanet_archive_mappings()
    except:
        param = {**Param.load().value, **{}}
        local_path = param["archive_mappings_csv"]
        label_df = load_exoplanet_archive_mappings(local_path)

    label_df_all = label_df.copy()
    label_df = label_df.dropna(subset=[PS_NAME, PC_NAME])
    label_df = label_df[~label_df[PS_NAME].str.contains("systemref")]
    label_ps = label_df[PS_NAME].str.strip()
    label_pc = label_df[PC_NAME].str.strip()

    for lps, lpc in zip(label_ps, label_pc):
        # Some columns are missing even if in CSV
        if lps not in MISSING_COLS:
            try:
                ps_tbl.rename_column(lps, lpc)
            except KeyError:
                warn(f"Column {lps} was not found in PS table", RuntimeWarning)
    # Ref time keys are mismatched in CSV so one will be missing, need to update
    tper_ref = "pl_orbtper_systemref"
    tranmid_ref = "pl_tranmid_systemref"
    tsys_ref = "pl_tsystemref"
    ps_tbl[tranmid_ref] = ps_tbl[tsys_ref]
    ps_tbl[tper_ref] = ps_tbl[tsys_ref]

    # Format references to match composite table format (masterfile default)
    if verbose:
        print("Formatting references to match composite table format...", end="")

    reflink_labels = label_df_all.dropna(subset=PC_NAME)[PC_NAME]
    reflink_labels = reflink_labels.str.strip()
    reflink_labels= reflink_labels[reflink_labels.str.endswith("reflink")]
    for rlab in reflink_labels:

        plab = rlab.replace("_reflink", "")
        if isinstance(ps_tbl[plab][0], str) or ps_tbl[plab].dtype.kind in ["U", "S"]:
            nmask = ps_tbl[plab] == ""
        else:
            nmask = np.isnan(ps_tbl[plab])


        if rlab.startswith("st_"):
            new_refs = ps_tbl["st_refname"]
        elif rlab.startswith(("sy_", "ra_")):
            new_refs = ps_tbl["sy_refname"]
        elif rlab.startswith("pl_"):
            new_refs = ps_tbl["pl_refname"]
        else:
            warn(f"Reference {rlab} has no known match. Setting to ''.")
            new_refs = ""

        ps_tbl[rlab] = MaskedColumn(np.where(nmask, "", new_refs))

    if verbose:
        print("Done")

    return ps_tbl


def compose_from_ps(
        ps_tbl,
        sort_keys: Union[str, List[str]],
        verbose: bool = False,
    ):
    # This creates a composite table, but using full rows instead of using the most precise
    # for each parameter individually as (I think) done for NASA archive.

    default_mask = ps_tbl["default_flag"] == 1
    extended = ps_tbl[~default_mask]
    ps_tbl = ps_tbl[default_mask]

    # Add estimate of transit mid time error as of today
    extended.estim_ephemeride_err()

    # Add estimate of transit mid time error as of today
    extended.estim_ephemeride_err(ephemeride="pl_orbtper")

    # Sort to choose the reference accordingly
    extended.sort(sort_keys)

    # Separate bet
    # Group by planet's name
    grouped = extended.group_by(["pl_name"])

    # Complete new table with extended table
    if verbose:
        print(
            "Completing database with extended database (may take some time)...",
            end="",
        )

    # Fill all the values we can until nothing left or ps is completely filled
    while len(grouped) > 1 and ps_tbl.has_masked_values:

        # Groupe by planet name
        grouped = grouped.group_by(["pl_name"])

        # "index" gives the first entry for each planet
        # (i.e. the star of each group where table is grouped by pl_name)
        index = grouped.groups.indices[:-1]

        # Add to master table
        # NOTE: This assumes that reflink will be masked if the value is masked.
        # Usually true, but would not hurt to check it somewhere
        ps_tbl = ps_tbl.complete(
            grouped[index], "pl_name", add_col=False, verbose=False
        )

        # Remove from the main table
        grouped.remove_rows(index)

    if verbose:
        print("Done")

    return ps_tbl



from collections import OrderedDict
from warnings import warn

from astropy.table import join
import astropy.table as table
from astropy.table.operations import _join, _merge_table_meta
from astropy.units import Unit
import numpy as np
from astropy.coordinates import SkyCoord
from .exceptions import MultipleResultsWarning, NotFoundWarning, NotFoundError


class MaskedColumn(table.MaskedColumn):

    def find(self, sub, start=0, end=None):
        if isinstance(self, (MaskedColumn)):
            str_array = np.array(self, dtype=str)
            index = np.core.defchararray.find(str_array, sub, start=start, end=end) != -1
            return np.where(index)[0], str_array[index]

        else:
            return NotImplemented

    def to_array(self, units=None):
        """
        Returns the columns as a MaskedArray.
        If units are specified, convert to good units
        before returning the array.
        """
        if units is None:
            return self.data.copy()
        else:
            # Convert to quantity array
            data = self.quantity
            # Need to stored masked values
            # (quantity arrays don't deal with masks)
            mask = self.mask

            # Make sure it has the good units
            data = data.to(units)

            # Convert to masked array
            return np.ma.array(data.value, mask=mask)


class Column(table.Column):

    def find(self, sub, start=0, end=None):
        if isinstance(self, (Column)):
            str_array = np.array(self, dtype=str)
            index = np.core.defchararray.find(str_array, sub, start=start, end=end) != -1
            return np.where(index)[0], str_array[index]

        else:
            return NotImplemented

    def to_array(self, units=None):
        """
        Returns the columns as an array.
        If units are specified, convert to good units
        before returning the array.
        """
        if units is None:
            return self.data.copy()
        else:
            # Convert to quantity array
            data = self.quantity

            # Make sure it has the good units
            data = data.to(units)

            # Convert to array
            return np.array(data.value)


class Table(table.Table):

    # Redefine class attributes (if not, the originals would be taken)
    Column = Column
    MaskedColumn = MaskedColumn

    # Set attributes
    main_col = None  # Default column used to order
    log = []  # Save output when using insert_value method

    def __init__(self, *args, masked=True, **kwargs):

        super().__init__(*args, masked=masked, **kwargs)

        if self.masked:
            self.nan_to_mask()

    # New methods
    def rename_columns(self, old, new):

        for ko, kn in zip(old, new):
            self.rename_column(ko, kn)

    def nan_to_mask(self):
        """
        Replace nan by masked array
        """
        if self.masked:
            for k in self.keys():
                if not isinstance(self[k], SkyCoord) and self[k].dtype == float:
                    self[k].mask = np.isnan(self[k]) | self[k].mask
        else:
            raise TypeError("Input must be a Masked Table." +
                            "\n \t Set its mask to True before calling" +
                            " (example: t = Table(t, masked=True, copy=False)).")

    def by_pl_name(self, *plName, name_key=None):
        """
        Return the complete line of a given planet name (plName)
        """
        position = self.get_index(*plName, name_key=name_key)
        position = np.array(position)

        is_valid = (position >= 0)

        if is_valid.any():
            out = self[position[is_valid]]
        else:
            raise NotFoundError(*plName)

        return out

    def by_plName(self, *args, **kwargs):
        """
        Old name of by_pl_name. Just an alias
        """

        return self.by_pl_name(*args, **kwargs)

    def get_index(self, *plName, name_key=None):
        '''
        Return the lines index where plName are located for the column given by name_key
        name_key default is given by main_col attribute of the object
        '''
        name_key = name_key or self.main_col

        position = []
        for pl in plName:
            try:
                position.append(int(self[name_key].find(pl)[0]))

            except TypeError:
                values = self[name_key].find(pl)[1]
                if len(values) > 0:
                    warn(MultipleResultsWarning(pl, values))
                else:
                    warn(NotFoundWarning(pl))
                position.append(-1)

        return position

    def set_main_col(self, colname=None, extension='_temp'):
        '''
        Set self.main_col and assign it to the first column.
        If colname is None, simply assign self.main_col to the
        first column.
        '''
        if self.main_col is None:
            self.main_col = colname
        elif colname is None:
            colname = self.main_col

        colname_temp = colname+extension
        self.rename_column(colname, colname_temp)
        self.add_column(self[colname_temp], name=colname, index=0)
        self.remove_column(colname_temp)

    def correct_units(self, badunits=['degrees', 'days', 'hours','jovMass', 'mags'],
                     gunits=['degree', 'day', 'hour','jupiterMass', 'mag'], verbose=True,
                     debug=False):
        '''
        Correct columns units for astropy units
        '''
        text_frame = "Column {} corrected for '{}' unit (previous was '{}')"

        for col in self.colnames:
            if debug:
                print(col, self[col].unit)

            # Skip skycoord: no unit attribute
            if isinstance(self[col], SkyCoord):
                continue

            # Search for bad units
            # TODO: check if unit is in list instead of loop and use dict
            # to replace
            for bunit, gunit in zip(badunits, gunits):
                if self[col].unit == bunit:
                    self[col].unit = gunit

                    # Message and log it
                    self.log.append(
                        text_frame.format(col, self[col].unit, bunit))
                    if verbose:
                        print(self.log[-1])
                    print(self.log[-1])

    def cols_2_qarr(self, *keys):
        '''
        Returns columns given in input as astropy q_arrays
        '''

        warn(DeprecationWarning(
            "This method will be removed in future versions. Do not use it."))

        out = []
        for k in keys:
            try:
                out.append(np.ma.array(self[k].data) * self[k].unit)
            except TypeError:
                out.append(np.ma.array(self[k].data))

        return tuple(out)

    def set_units(self, units, cols=None):
        '''
        Assign units to columns.
        units:
          list of units (str or astropy units) to be assign
        cols:
          list of columns names (str).
          If None is given, it takes all the keys, so Table.keys() as default
        '''

        if not cols:
            cols = self.keys()

        for col, u in zip(cols, units):
            self[col].unit = u

    def new_value(self, plName, col, value):

        names = np.array(self[self.main_col], dtype=str)
        position = np.where(names == plName)[0]

        self[col][position] = value

    def complete(self, right, key=None, join_type='left',
                 add_col=True, metadata_conflicts='warn',
                 verbose=True, debug=False, **kwargs):
        """
        Add every missing data in self if present in right.

        join_type : 'inner': do not add new rows
                    'outer': add new rows if not present in self
        add_col: add new colums from right
        """

        key = key or self.main_col

        # Try converting inputs to Table as needed
        if not isinstance(right, Table):
            right = Table(right)

        try:
            out = self._complete(right, key=key, join_type=join_type,
                                 add_col=add_col, verbose=verbose, debug=debug)
        except:
            warn("Custom table completion failed, trying default astropy join.")
            # NOTE: This seemd to break when I tested it quickly,
            # but I needed masking anyway so I did not get to the bottom of the problem.
            # - Thomas
            col_name_map = OrderedDict()
            out = _join(
                self,
                right,
                join_type=join_type,
                col_name_map=col_name_map,
                keys=key,
                **kwargs
            )

        # Merge the column and table meta data. Table subclasses might override
        # these methods for custom merge behavior.
        _merge_table_meta(out, [self, right], metadata_conflicts=metadata_conflicts)

        return out

    def _complete(self, right, key=None, join_type='left', add_col=True,
                  verbose=True, debug=False):

        if not key:
            raise ValueError('key is empty')

        # Save shared columns without "key"
        cols = intersection(self.keys(), right.keys())
        cols.remove(key)

        # Join tables
        join_t = join(self, right, join_type=join_type, keys=key)

        # Complete masked values of "self" if available in "right"
        for col in cols:

            # Add eventually a condition to check units!

            # Names of joined columns (default from join())
            col1, col2 = col + '_1', col + '_2'

            # Index of masked in "self" and not masked in "right"
            index = join_t[col1].mask & ~join_t[col2].mask

            # Reassign value
            join_t[col1].unshare_mask()
            join_t[col1][index] = join_t[col2][index]

            # Remove 2nd column and rename to original
            join_t[col1].name = col
            del join_t[col2]

        # Remove added columns from "right" if not wanted
        supp_cols = difference(right.keys(), self.keys())
        if debug: print(supp_cols)

        if not add_col and supp_cols:
            if verbose:
                print('remove non shared columns from second table')
            join_t.remove_columns(supp_cols)

        return join_t

    def complete_cols(self, col_in, col_out, name_key=None):
        '''
        Use a column from table to complete another column.
        Input:
            col_in: list of names of columns to use (list of str)
            col_out: list of names of columns to complete (list of str)
        '''
        # Take default col if none is given
        name_key = name_key or self.main_col

        # Def table with cols to use and rename it to cols to complete
        temp_table = Table(self[[name_key] + col_in], masked=True)
        temp_table.nan_to_mask()
        temp_table.rename_columns(col_in, col_out)

        # Complete with the temp_table
        return self.complete(temp_table, key=name_key)

    def add_calc_col(self, fct, *args, f_args=(), f_kwargs={}, col_keys=[], **kwargs):
        '''
        Add new column wich is the result of fct(table[col_keys], *f_args, **f_kwargs)

        args and kwargs are passed to MaskedColumn instantiation

        '''

        # Build tuple of columns inputs to fct and add to f_args
        cols = ()
        for key in col_keys:
            cols += (self[key],)
        f_args = cols + f_args

        # Define column and add it
        col = MaskedColumn(*args, data=fct(*f_args, **f_kwargs), **kwargs)
        self.add_column(col)

    def check_col_units(self, colname):

        col_units = self[colname].unit
        try:  # Check if col_units valid
            1. * Unit(col_units)
        except TypeError:
            print('Column has no units (unit = None)')
        except:  # Enter valid unit and refresh all table
            print("Column units '{}' are not".format(col_units) +
                          ' recognized by astropy.\n')
            print("Error message from astropy:")
            print_unit_error(str(col_units))
            print("-------------------------")
            gunit = input('***** Please enter the corresponding unit'
                          + ' recognized by astropy unit: ')
            self.correct_units(badunits=[str(col_units)], gunits=[gunit])


def difference(left, right):
    if isinstance(left, list) and isinstance(right, list):
        return list(set(left) - set(right))
    else:
        return NotImplemented


def intersection(tbl, other):
    if isinstance(tbl, list):
        return list(set(tbl).intersection(other))
    else:
        return NotImplemented


def print_unit_error(str_unit):

    try:
        Unit(str_unit)
    except ValueError as e:
        print(e)



from exofile.archive import GoogleSheet
from exofile.utils import migrate_table

old_key = "14Trm-AQ2eOphfwqJYrevnDrNVk56E-aH8yvRHQLjzWg"
gs_tbl = GoogleSheet.query(old_key)
gs_ref = GoogleSheet.query(old_key, sheet_name="Ref")

for rc in gs_ref.colnames:
    gs_ref[rc] = gs_ref[rc].astype("str")
    gs_ref[rc].unit = None

new_gs = migrate_table(gs_tbl, gs_ref)
columns = new_gs.colnames

# Bring back to GS col names (units in bracket)
for col in columns:
    new_name = f'{new_gs[col].name} [{new_gs[col].unit or None}]'
    new_gs.rename_column(col, new_name)

new_gs = GoogleSheet(new_gs, copy=False, masked=True)
new_gs.nan_to_mask()
# Write locally to CSV, then users can copy past with spreadsheet editor
new_gs.write("new_google_sheet.csv")



from collections import OrderedDict
from os import remove
from os.path import abspath, dirname
from warnings import warn

import yaml


# --------------------------------
# Useful functions
# --------------------------------

def get_module_path(file):

    dir_path = abspath(file)
    dir_path = dirname(dir_path) + '/'

    return dir_path

# ----------------------------------
# Functions for param
# ----------------------------------
class Param():

    default_file = get_module_path(__file__) + 'param.yaml'

    def __init__(self, value, description=None):

        # Make sure they are OrderedDict
        value = OrderedDict(value)

        if description is None:
            self.value = OrderedDict()
            self.description = OrderedDict()
            for key in value:
                try:
                    self[key] = value[key]
                except TypeError:
                    raise
        else:
            description = OrderedDict(description)

            self.value = value
            self.description = description

    @classmethod
    def load(cls, filename=None, raise_err=False, **kwargs):

        kwargs = {'Loader':yaml.Loader, **kwargs}

        if filename is None:
            output = cls._load_default(raise_err, **kwargs)
            return cls(output)

        try:
            with open(filename) as f:
                output = yaml.load(f, **kwargs)
        except FileNotFoundError as e:
            message = str(e) + '. Taking default param instead.'
            warn(message)
            output = cls._load_default(raise_err, **kwargs)

        return cls(output)

    @classmethod
    def _load_default(cls, raise_err, **kwargs):

        try:
            with open(cls.default_file) as f:
                    output = yaml.load(f, **kwargs)
        except FileNotFoundError as e:
            if raise_err:
                message = str(e) \
                   + ". You need to configurate the exofile. \n"  \
                   + "Example:  \n"  \
                   + ">>> from exofile.config import edit_param  \n" \
                   + ">>> edit_param(sheet_key='the_key_to_the_google_sheet',) \n"  \
                   + ">>> edit_param(url='url_to_exofile',"  \
                   + " url_ref='url_to_exofile_references')"
                raise FileNotFoundError(message)
            else:
                configurate()
                with open(cls.default_file) as f:
                        output = yaml.load(f, **kwargs)
        return output

    def dump(self, filename=None, **kwargs):

        if filename is None:
            filename = self.default_file

        with open(filename, 'w') as f:
            yaml.dump(self.to_dict(), f, **kwargs)


    def to_dict(self):

        out = OrderedDict()
        for key in self.value:
            out[key] = {'Value':self.value[key],
                        'Description':self.description[key]}
        return out

    def __repr__(self):

        return yaml.dump(self.to_dict())

    def __eq__(self, other):

        return self.value == other.value


    def __getitem__(self, key):

        return {'Value':self.value[key], 'Description':self.description[key]}

    def __setitem__(self, key, value):

        if isinstance(value, dict):
            self.value[key] = value['Value']
            self.description[key] = value['Description']
        elif isinstance(value, list):
            self.value[key] = value[0]
            self.description[key] = value[1]
        else:
            self.value[key] = value
            self.description[key] = None

    def keys(self):

        return self.value.keys()

    def __iter__(self):

        yield from self.keys()



def edit_param(**param_kwargs):

    # Load param default file (param.yaml in module)
    param = Param.load()

    # Convert into Param object
    param_kwargs = Param(param_kwargs)

    # Take default description if not given
    for key in param_kwargs:
        desc = param_kwargs.description
        if desc[key] is None:
            try:
                desc[key] = param.description[key]
            except KeyError as _:
                message = "No description found for '{}'"
                warn(message.format(key))

    # Combine Param objects
    param = Param({**param, **param_kwargs})

    # Save to param.yaml in module directory
    param.dump()


def configurate():
    '''
    Copy default_param to param
    '''

    # Get librairy dir
    dir_path = get_module_path(__file__)

    param = Param.load(dir_path + 'default_param.yaml')

    with open(dir_path + 'default_param.yaml', 'r') as f:
            param = f.readlines()

    try:
        with open(dir_path + 'param.yaml', 'x') as f:
            f.writelines(param)
    except FileExistsError as e:
        message = str(e) + '. Run exofile.config.reset_param() to delete the file.'
        raise FileExistsError(message)

def reset_param():
    '''
    Delete param.yaml (located in the librairy dir)
    '''

    # Get librairy dir
    dir_path = get_module_path(__file__)

    # Remove the file
    remove(dir_path + 'param.yaml')